--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

package Games.Keen is

private

    type Keen_Game is new Game with
        record
            dialogId   : Integer := Integer'First;    -- current open dialog
            mapWorld   : A_World;
            playingMap : Boolean := True;    -- playing the overhead map
        end record;

    procedure Add_Event_Listeners( this : access Keen_Game );

    procedure Delete( this : in out Keen_Game );

    procedure End_Game( this : access Keen_Game; interrupted : Boolean );

    procedure Handle_Event( this : access Keen_Game;
                            evt  : in out A_Event;
                            resp : out Response_Type );
    pragma Precondition( evt /= null );

    -- Called when the Load_World event is received, to load the next level. If
    -- the world name to load is an empty string, the current level will be
    -- exited, marked as complete, and the overhead map will be loaded. No
    -- exception will be raised on error; the error message will be put into the
    -- response 'resp'.
    procedure Handle_Load_World( this : access Keen_Game;
                                 evt  : not null A_Load_World_Event;
                                 resp : out Response_Type );

    procedure New_Game( this : access Keen_Game );

    procedure Pause( this : access Keen_Game; enabled : Boolean );

    procedure Remove_Event_Listeners( this : access Keen_Game );

end Games.Keen;
