--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Tiles.Keen;                        use Tiles.Keen;

package Worlds.Keen is

    type Keen_World is new World_Object with private;
    type A_Keen_World is access all Keen_World'Class;

    -- Returns the 'action' attribute of the tile at the given location. The
    -- dominant action of the tiles in each layer at 'x,y' will be returned.
    function Get_Action( this : not null access Keen_World'Class;
                         x, y : Integer ) return Action_Type;

private

    type Keen_World is new World_Object with
        record
            -- the following fields are streamed
            introduction : Unbounded_String;
        end record;

    -- Constructs the world without a map or library name; used for streaming.
    procedure Construct( this : access Keen_World );

    -- Constructs a new empty world with the given arguments
    procedure Construct( this    : access Keen_World;
                         width,
                         height  : Positive;
                         libName,
                         domain  : String );
    pragma Precondition( libName'Length > 0 );
    pragma Precondition( domain'Length > 0 );

    function Object_Input( stream : access Root_Stream_Type'Class ) return Keen_World;
    for Keen_World'Input use Object_Input;

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Keen_World );
    for Keen_World'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Keen_World );
    for Keen_World'Write use Object_Write;

    procedure Queue_Load_Events( this : access Keen_World );

    -- extended keen properties:
    -- "introduction" : the introduction text
    procedure Set_Property( this : access Keen_World; name, value : String );
    pragma Precondition( name'Length > 0 );

end Worlds.Keen;
